#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#

from ibvpy.fets import FETSEval

#-----------------------------------------------------------------------------------
# FETS3D - Base class for 3D elements 
#-----------------------------------------------------------------------------------

class FETS3D(FETSEval):
    '''Base class for 3D elements.
    '''

