#-------------------------------------------------------------------------------
#
# Copyright (c) 2020, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.
#
# Thanks for using BMCS open source!
#

from ibvpy.fets import FETSEval

class FETS1D5(FETSEval):
    '''Base class for 2D elements.
    '''

